from logging import getLogger
from .data import Data, DataType
from typing import TypeVar

logger = getLogger(__name__)
TRepository = TypeVar("Repository")


class RawData(Data):
    def __init__(self, content: bytes, repository: TRepository = None):
        super().__init__(repository, DataType.RAW)
        self.content_ = content

    @classmethod
    def load(cls, repository: TRepository) -> "RawData":
        return repository.load(RawData.create)

    @classmethod
    def create(cls, raw_data: bytes, repository: TRepository = None) -> "RawData":
        return RawData(raw_data, repository)

    @property
    def content(self) -> bytes:
        return self.content_

    def __str__(self) -> str:
        return f"RawData: {len(self.content)}bytes, b'{self.content[:16]}...'"



from logging import getLogger
from aksdp.graph import Graph, TaskStatus
import time
import zlib
import base64

logger = getLogger(__name__)


class PlantUML:
    @classmethod
    def graph_to_plantuml(cls, graph: Graph) -> str:
        components = [f"[{t.task.__class__.__name__}]" for t in graph.graph]

        dependencies = []
        for g in graph.graph:
            _to = g.task.__class__.__name__
            _to_datakeys = g.task.input_datakeys()

            if g.dependencies:
                for t in g.dependencies:
                    _from = t.task.__class__.__name__
                    _from_datakeys = t.task.output_datakeys()

                    # I/Oの datakey が宣言されていれば注釈追加
                    datakeys = list(set(_to_datakeys) & set(_from_datakeys))
                    datamemo = ""
                    if datakeys:
                        datamemo = f" : {','.join(datakeys)}"

                    dependencies.append(f"{_from} --> {_to}{datamemo}")

        _components = "\n".join(components)
        _dependencies = "\n".join(dependencies)
        uml = f"""
@startuml
{_components}
{_dependencies}
@enduml
"""
        return uml

    @classmethod
    def graph_to_plantuml_textenc(cls, graph: Graph) -> str:
        s = PlantUML.graph_to_plantuml(graph)
        compressed = zlib.compress(s.encode("utf-8"))
        encoded = base64.b64encode(compressed[2:-4])

        puml_base64chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_".encode("utf-8")
        std_base64chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".encode("utf-8")

        s = encoded.translate(bytes.maketrans(std_base64chars, puml_base64chars)).decode("utf-8")

        return s

    @classmethod
    def graph_to_url(cls, graph: Graph, base_url: str = "http://www.plantuml.com/plantuml/png/") -> str:
        return f"{base_url}{PlantUML.graph_to_plantuml_textenc(graph)}"



import boto3
import tempfile
from logging import getLogger
from .localfile_repository import LocalFileRepository
from pathlib import Path
from aksdp.data import Data
from urllib.parse import urlparse

logger = getLogger(__name__)


class S3FileRepository(LocalFileRepository):
    """S3上のファイルへのアクセス
    LocalFileの機能に読み込み・保存の前後にS3への転送を挟むだけ。
    """
    def __init__(self, access_key_id: str, secret_access_key: str, url: str):
        o = urlparse(url, allow_fragments=False)

        self.tempdir = tempfile.TemporaryDirectory()
        self.path = Path(self.tempdir.name) / Path(o.path).name
        self.s3_url = url
        self.s3_bucket = o.netloc
        self.s3_key = o.path.lstrip("/")

        logger.debug(
            f"setup S3FileRepository... tempdir={self.tempdir.name}, path={str(self.path)}, s3_bucket={self.s3_bucket}"
            f" s3_key={self.s3_key}"
        )

        self.s3client = boto3.Session(aws_access_key_id=access_key_id, aws_secret_access_key=secret_access_key,).client(
            "s3"
        )

        self._upload_extra_args = {}

    def load(self, ctor):
        logger.debug(f"s3 download {self.s3_url} -> {self.path.resolve()}")
        with open(self.path.resolve(), "wb") as f:
            self.s3client.download_fileobj(self.s3_bucket, self.s3_key, f)
        return super().load(ctor)

    def save(self, data: Data):
        super().save(data)

        logger.debug(f"s3 upload {str(self.path.resolve())} -> {self.s3_url}")
        self.s3client.upload_file(str(self.path.resolve()), self.s3_bucket, self.s3_key,
            ExtraArgs=self.upload_extra_args)

    @property
    def upload_extra_args(self) -> dict:
        return self._upload_extra_args
    
    @upload_extra_args.setter
    def upload_extra_args(self, extra_args:dict):
        self._upload_extra_args = extra_args



from sqlalchemy.orm import scoped_session, sessionmaker
from logging import getLogger
from aksdp.data import Data, DataType
from .repository import Repository

logger = getLogger(__name__)


class SqlAlchemyRepository(Repository):
    Session = None

    def __init__(self, engine):
        session_factory = sessionmaker(bind=engine)
        Session = scoped_session(session_factory)
        self.session = Session()

    def __del__(self):
        """.dtor
        セッションのクローズ
        """
        if self.session:
            self.session.close()

    def save(self, data: Data):
        if data.data_type == DataType.SQLALCHEMY_MODEL:
            self.commit()
        else:
            raise ValueError(f"SqlAlchemyRepository.save() not support DataType {data.data_type}")

    def query(self, model_class):
        return self.session.query(model_class)

    def insert(self, model):
        self.session.add(model)
        return model

    def delete(self, model):
        self.session.delete(model)

    def commit(self):
        self.session.commit()

    def rollback(self):
        self.session.rollback()



from .data import Data as Data
from .data import DataType as DataType
from .dataframe_data import DataFrameData as DataFrameData
from .json_data import JsonData as JsonData
from .raw_data import RawData as RawData
from .sqlalchemy_model_data import SqlAlchemyModelData as SqlAlchemyModelData



from logging import getLogger
from .data import Data, DataType
import pandas as pd
from io import StringIO
from typing import TypeVar

logger = getLogger(__name__)
TRepository = TypeVar("Repository")


class DataFrameData(Data):
    def __init__(self, content: pd.DataFrame, repository: TRepository = None):
        super().__init__(repository, DataType.DATAFRAME)
        self.content_ = content

    @classmethod
    def load(cls, repository: TRepository) -> "DataFrameData":
        return repository.load(DataFrameData.create_from_csv)

    @classmethod
    def create_from_csv(cls, raw_data: bytes, repository: TRepository = None) -> "DataFrameData":
        sio = StringIO(raw_data.decode("utf-8"))
        df = pd.read_csv(sio)
        return DataFrameData(df, repository)

    @classmethod
    def create_from_df(cls, df: pd.DataFrame, repository: TRepository = None) -> "DataFrameData":
        return DataFrameData(df, repository)

    @property
    def content(self) -> pd.DataFrame:
        return self.content_

    def __str__(self) -> str:
        return f"DataFrameData:¥n{self.content.head()}"



from .data import Data, DataType
import pandas as pd
from logging import getLogger
from typing import TypeVar

logger = getLogger(__name__)
TRepository = TypeVar("Repository")


class SqlAlchemyModelData(Data):
    def __init__(self, repository: TRepository, model_class):
        super().__init__(repository, DataType.SQLALCHEMY_MODEL)
        self.model_class = model_class
        self.content_ = []

    def to_dataframe(self) -> pd.DataFrame:
        """モデルのリストを DataFrame に変換する"""
        columns = [c.name for c in self.model_class.__table__.columns]

        df = pd.DataFrame(columns=columns)
        for r in self.content:
            series_data = [getattr(r, c.name) for c in self.model_class.__table__.columns]
            series = pd.Series(series_data, index=df.columns)
            df = df.append(series, ignore_index=True)

        return df

    def update_dataframe(self, df: pd.DataFrame):
        """DataFrameの内容をモデルのリストに書き戻す"""
        append_entities = []

        def update_model(inst, row):
            for c in self.model_class.__table__.columns:
                setattr(inst, c.name, row[c.name])

        for index, row in df.iterrows():
            if index >= len(self.content):
                inst = self.model_class()
                append_entities.append(inst)
            else:
                inst = self.content[index]

            update_model(inst, row)

        # DataFrameから削除された行にマッチするモデルの削除
        indexes = list(df.index.values)
        for idx in reversed(range(0, len(self.content))):
            if idx not in indexes:
                m = self.content.pop(idx)
                self.repository.delete(m)

        # 増えた分をINSERT
        append_models = list(map(lambda x: self.repository.insert(x), append_entities))
        self.content.extend(append_models)

    def query(self, qf=None):
        q = self.repository.query(self.model_class)
        if qf:
            q = qf(q)
        self.content_ = list(q.all())

    @property
    def content(self) -> list:
        return self.content_

    def __str__(self) -> str:
        return f"SqlAlchemyModelData: {len(self.content)} records"



from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from aksdp.task import Task
from aksdp.dataset import DataSet
from aksdp.graph import Graph, TaskStatus
from typing import List
from enum import Enum
import time
from .graph import GraphTask


logger = getLogger(__name__)


class ConcurrentGraph(Graph):
    def __init__(self, executor=None):
        super().__init__()

        self.pool = executor
        if self.pool is None:
            self.pool = ThreadPoolExecutor()

    def _run(self, graph_task: GraphTask, input_ds: DataSet):
        # ProcessPoolを使用した場合、 GraphTask.run() 内でのアトリビュート更新が効かないので外から操作する
        graph_task.input_ds = input_ds
        graph_task.status = TaskStatus.RUNNING
        return self.pool.submit(graph_task.run, input_ds)

    def run(self, ds: DataSet = None):
        last_ds = ds
        features = []

        while self.is_all_completed() is False and self.abort is False:
            tasks = self.runnable_tasks()

            for t in tasks:
                input_ds = self._make_task_inputs(t)
                input_ds = ds if input_ds is None else input_ds
                features.append((t, self._run(t, input_ds)))

            # どれかの Feature が終わるまで待つ
            def all_feature_running(features):
                return all([f.running() for f in features])

            while all_feature_running([f[1] for f in features]):
                time.sleep(0.1)

            # 終了した features の result を吸い上げ、features を更新
            _next_featrues = []
            for f in features:
                if f[1].done():
                    # ProcessPoolを使用した場合、 GraphTask.run() 内でのアトリビュート更新が効かないので外から操作する
                    try:
                        last_ds = f[0].output_ds = f[1].result()
                        f[0].status = TaskStatus.COMPLETED
                    except BaseException as e:
                        if not self._handle_error(f[0], f[0].input_ds, e):
                            raise
                else:
                    _next_featrues.append(f)
            features = _next_featrues

        return last_ds



from logging import getLogger
from .repository import Repository
from pathlib import Path
from aksdp.data import Data, DataType
import json

logger = getLogger(__name__)


class LocalFileRepository(Repository):
    def __init__(self, path: Path):
        super().__init__()
        self.path = path

    def save(self, data: Data):
        if data.data_type == DataType.RAW:
            self.path.write_bytes(data.content)
        elif data.data_type == DataType.JSON:
            json_str = json.dumps(data.content, ensure_ascii=False)
            self.path.write_bytes(json_str.encode("utf-8"))
        elif data.data_type == DataType.DATAFRAME:
            data.content.to_csv(str(self.path), index=False)
        else:
            raise ValueError("LocalFileRepository.save() not support DataType {data.data_type}")

    def load(self, ctor):
        raw_data = self.path.read_bytes()
        return ctor(raw_data, self)



from logging import getLogger
from enum import Enum
from typing import TypeVar, Generic

logger = getLogger(__name__)
TRepository = TypeVar("Repository")


class DataType(Enum):
    RAW = 0
    DATAFRAME = 1
    JSON = 2
    SQLALCHEMY_MODEL = 3


class Data:
    def __init__(self, repository: TRepository, data_type: DataType = DataType.RAW):
        self.repository = repository
        self.data_type_ = data_type

    def save(self):
        if self.repository:
            self.repository.save(self)

    @property
    def data_type(self) -> DataType:
        return self.data_type_



from concurrent.futures import ThreadPoolExecutor, as_completed
from logging import getLogger
from aksdp.task import Task
from aksdp.dataset import DataSet
from typing import List, Callable
from enum import Enum

logger = getLogger(__name__)


class TaskStatus(Enum):
    INIT = 0
    RUNNING = 1
    COMPLETED = 2
    ERROR = 3


class GraphTask(object):
    def __init__(self, task: Task, dependencies: List["GraphTask"]):
        """.ctor

        Args:
            task (Task): 実行タスク
            dependencies (List): 依存タスク
        """
        self.task = task
        self.dependencies = dependencies
        self.status = TaskStatus.INIT
        self.input_ds = None
        self.output_ds = None
        self._pre_run_hook = self.empty_hook
        self._post_run_hook = self.empty_hook

    @property
    def pre_run_hook(self) -> Callable:
        """実行前フックの取得

        Returns:
            Callable: 実行前フック関数
        """
        return self._pre_run_hook

    @pre_run_hook.setter
    def pre_run_hook(self, v: Callable):
        """実行前フックの設定

        Args:
            v (Callable): 実行前フック関数
        """
        self._pre_run_hook = v

    @property
    def post_run_hook(self) -> Callable:
        """実行後フックの取得

        Returns:
            Callable: 実行後フック関数
        """
        return self._post_run_hook

    @post_run_hook.setter
    def post_run_hook(self, v: Callable):
        """実行後フックの設定

        Args:
            v (Callable): 実行後フック関数
        """
        self._post_run_hook = v

    def empty_hook(self, ds: DataSet):
        """デフォルトの空フック

        Args:
            ds (DataSet): データセット
        """
        pass

    def is_runnable(self) -> bool:
        if self.status != TaskStatus.INIT:
            return False

        if self.dependencies is None:
            return True

        return all([gt.status == TaskStatus.COMPLETED for gt in self.dependencies])

    def run(self, ds: DataSet = None) -> DataSet:
        try:
            self.input_ds = ds
            logger.debug(f"task({self.task.__class__.__name__}) started.")
            logger.debug(f"  input_ds = {str(ds)}")
            self.pre_run_hook(ds)
            output_ds = self.task.gmain(ds)
            self.post_run_hook(output_ds)

            logger.debug(f"task({self.task.__class__.__name__}) completed. (elapse={self.task.elapsed_time:.3f}s)")
            logger.debug(f"  output_ds = {str(output_ds)}")
            self.output_ds = output_ds
        except BaseException as e:
            logger.error(f"task({self.task.__class__.__name__}) failed. {str(e)}")
            raise
        return self.output_ds


class Graph:
    def __init__(self):
        self.graph = []
        self.error_handlers = []
        self.abort = False

    def append(self, task: Task, dependencies: List[GraphTask] = []) -> GraphTask:
        gt = GraphTask(task, dependencies)
        self.graph.append(gt)
        return gt

    def add_error_handler(self, cls, fn):
        self.error_handlers.append((cls, fn))

    def runnable_tasks(self) -> List[GraphTask]:
        return [g for g in self.graph if g.is_runnable()]

    def _run(self, graph_task, input_ds):
        try:
            graph_task.status = TaskStatus.RUNNING
            r = graph_task.run(input_ds)
            graph_task.status = TaskStatus.COMPLETED
            return r
        except BaseException as e:
            if not self._handle_error(graph_task, input_ds, e):
                raise

    def _handle_error(self, graph_task: GraphTask, input_ds: DataSet, e: BaseException):
        graph_task.status = TaskStatus.ERROR
        for eh in self.error_handlers:
            if isinstance(e, eh[0]):
                self.abort = True
                eh[1](e, input_ds)
                return True

        return False

    def _make_task_inputs(self, graph_task):
        if not graph_task.dependencies:
            return None

        ds = DataSet()
        for d in graph_task.dependencies:
            ds.merge(d.output_ds)
        return ds

    def run(self, ds: DataSet = None) -> DataSet:
        last_ds = ds
        while self.runnable_tasks() and self.abort is False:
            t = self.runnable_tasks()[0]

            input_ds = self._make_task_inputs(t)
            input_ds = ds if input_ds is None else input_ds

            last_ds = self._run(t, input_ds)
        return last_ds

    def is_all_completed(self) -> bool:
        return all([gt.status == TaskStatus.COMPLETED for gt in self.graph])

    def autoresolve_dependencies(self):
        """グラフ依存関係の自動解決
        """
        logger.info("run task dependencies auto resolver...")

        def _find_datakey_provider(key: str) -> List[GraphTask]:
            return [gt for gt in self.graph if key in gt.task.output_datakeys()]

        for gt in self.graph:
            deps = set()
            for ik in gt.task.input_datakeys():
                prv = _find_datakey_provider(ik)
                prv_str = ",".join([gt.task.__class__.__name__ for gt in prv])

                if len(prv) > 1:
                    msg = f"data({ik}) provide from multiple task({prv_str})"
                    logger.error(msg)
                    raise ValueError(msg)

                if len(prv) == 0:
                    msg = f"task provide data({ik}) not found."
                    logger.info(msg)

                logger.debug(f"{gt.task.__class__.__name__} : depend on {prv[0].task.__class__.__name__} by data({ik})")
                deps.add(prv[0])

            gt.dependencies = list(deps)

        logger.info("task autoresolve completed.")
